/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var juju = { components: {} }; // eslint-disable-line no-unused-vars

chai.config.includeStack = true;
chai.config.truncateThreshold = 0;

describe('BooleanConfig', function () {

  beforeAll(function (done) {
    // By loading this file it adds the component to the juju components.
    YUI().use('boolean-config', function () {
      done();
    });
  });

  it('renders a checked input based on config prop', function () {
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: true,
      label: 'Test',
      option: option }));
    var expected = React.createElement(
      'div',
      { className: 'boolean-config' },
      React.createElement(
        'div',
        { className: 'boolean-config--title' },
        'Test'
      ),
      React.createElement(
        'div',
        { className: 'boolean-config--toggle' },
        React.createElement('input', {
          type: 'checkbox',
          id: option.key,
          onClick: output.props.children[1].props.children[0].props.onClick,
          onChange: output.props.children[1].props.children[0].props.onChange,
          checked: true,
          className: 'boolean-config--input' }),
        React.createElement(
          'label',
          {
            htmlFor: option.key,
            className: 'boolean-config--label' },
          React.createElement('div', { className: 'boolean-config--handle' })
        )
      ),
      React.createElement('div', { className: 'boolean-config--description',
        dangerouslySetInnerHTML: { __html: option.description } })
    );
    assert.deepEqual(output, expected);
  });

  it('renders an unchecked input based on config prop', function () {
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: false,
      label: 'Test',
      option: option }));
    assert.deepEqual(output, React.createElement(
      'div',
      { className: 'boolean-config' },
      React.createElement(
        'div',
        { className: 'boolean-config--title' },
        'Test'
      ),
      React.createElement(
        'div',
        { className: 'boolean-config--toggle' },
        React.createElement('input', {
          type: 'checkbox',
          id: option.key,
          onClick: output.props.children[1].props.children[0].props.onClick,
          onChange: output.props.children[1].props.children[0].props.onChange,
          checked: false,
          className: 'boolean-config--input' }),
        React.createElement(
          'label',
          {
            htmlFor: option.key,
            className: 'boolean-config--label' },
          React.createElement('div', { className: 'boolean-config--handle' })
        )
      ),
      React.createElement('div', { className: 'boolean-config--description',
        dangerouslySetInnerHTML: { __html: option.description } })
    ));
  });

  it('supports string boolean config props (true)', function () {
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: 'True',
      label: 'Test',
      option: option }));
    assert.deepEqual(output, React.createElement(
      'div',
      { className: 'boolean-config' },
      React.createElement(
        'div',
        { className: 'boolean-config--title' },
        'Test'
      ),
      React.createElement(
        'div',
        { className: 'boolean-config--toggle' },
        React.createElement('input', {
          type: 'checkbox',
          id: option.key,
          onClick: output.props.children[1].props.children[0].props.onClick,
          onChange: output.props.children[1].props.children[0].props.onChange,
          checked: true,
          className: 'boolean-config--input' }),
        React.createElement(
          'label',
          {
            htmlFor: option.key,
            className: 'boolean-config--label' },
          React.createElement('div', { className: 'boolean-config--handle' })
        )
      ),
      React.createElement('div', { className: 'boolean-config--description',
        dangerouslySetInnerHTML: { __html: option.description } })
    ));
  });

  it('supports string boolean config props (false)', function () {
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: 'False',
      label: 'Test',
      option: option }));
    assert.deepEqual(output, React.createElement(
      'div',
      { className: 'boolean-config' },
      React.createElement(
        'div',
        { className: 'boolean-config--title' },
        'Test'
      ),
      React.createElement(
        'div',
        { className: 'boolean-config--toggle' },
        React.createElement('input', {
          type: 'checkbox',
          id: option.key,
          onClick: output.props.children[1].props.children[0].props.onClick,
          onChange: output.props.children[1].props.children[0].props.onChange,
          checked: false,
          className: 'boolean-config--input' }),
        React.createElement(
          'label',
          {
            htmlFor: option.key,
            className: 'boolean-config--label' },
          React.createElement('div', { className: 'boolean-config--handle' })
        )
      ),
      React.createElement('div', { className: 'boolean-config--description',
        dangerouslySetInnerHTML: { __html: option.description } })
    ));
  });

  it('can call an onChange method if supplied', function () {
    var onChange = sinon.stub();
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: 'False',
      label: 'Test',
      onChange: onChange,
      option: option }));
    output.props.children[1].props.children[0].props.onChange({
      target: {
        checked: true
      }
    });
    assert.equal(onChange.callCount, 1);
  });

  it('can update when new config is provided', function () {
    var option = {
      key: 'testcheck',
      description: 'it is a test config option'
    };
    var shallowRenderer = jsTestUtils.shallowRender(React.createElement(juju.components.BooleanConfig, {
      config: true,
      label: 'Test',
      option: option }), true);
    var instance = shallowRenderer.getMountedInstance();
    assert.isTrue(instance.state.value);
    shallowRenderer.render(React.createElement(juju.components.BooleanConfig, {
      config: false,
      label: 'Test',
      option: option }));
    assert.isFalse(instance.state.value);
  });
});
