/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('entity-content', function () {

  juju.components.EntityContent = React.createClass({
    displayName: 'EntityContent',

    /* Define and validate the properites available on this component. */
    propTypes: {
      entityModel: React.PropTypes.object.isRequired,
      renderMarkdown: React.PropTypes.func.isRequired,
      getFile: React.PropTypes.func.isRequired,
      changeState: React.PropTypes.func.isRequired,
      pluralize: React.PropTypes.func.isRequired
    },

    /**
      Generate the list of configuration options.
       @method _generateOptionsList
      @param {Object} entityModel The entity model.
      @return {Object} The options markup.
    */
    _generateOptionsList: function _generateOptionsList(entityModel) {
      if (entityModel.get('entityType') === 'bundle') {
        return;
      }
      var options = entityModel.get('options');
      if (options) {
        var optionsList = [];
        Object.keys(options).forEach(function (name) {
          var option = options[name];
          option.name = name;
          optionsList.push(React.createElement(juju.components.EntityContentConfigOption, {
            key: name,
            option: option }));
        }, this);
        return React.createElement(
          'div',
          { id: 'configuration', className: 'row entity-content__configuration' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'twelve-col' },
              React.createElement(
                'h2',
                { className: 'entity-content__header' },
                'Configuration'
              ),
              React.createElement(
                'dl',
                null,
                optionsList
              )
            )
          )
        );
      }
    },

    /**
      Generates an HTML list from the supplied array.
       @method _generateList
      @param {Array} list The list of objects to markup.
      @param {Function} handler The click handler for each item.
      @return {Array} The list markup.
    */
    _generateList: function _generateList(list, handler) {
      return list.map(function (item) {
        return React.createElement(
          'li',
          { key: item },
          React.createElement(
            'a',
            { 'data-id': item, onClick: handler },
            item
          )
        );
      });
    },

    /**
      Generate the list of Tags if available.
       @method _generateTags
      @return {Array} The tags markup.
    */
    _generateTags: function _generateTags() {
      // Have to convert {0: 'database'} to ['database'].
      var tags = [],
          entityTags = this.props.entityModel.get('tags'),
          index;
      if (!entityTags) {
        return;
      }
      for (index in entityTags) {
        tags.push(entityTags[index]);
      }
      return React.createElement(
        'div',
        { className: 'four-col entity-content__metadata' },
        React.createElement(
          'h4',
          null,
          'Tags'
        ),
        React.createElement(
          'ul',
          null,
          this._generateList(tags, this._handleTagClick)
        )
      );
    },

    /**
      Handle clicks on tags.
       @method _handleTagClick
      @param {Object} e The event.
    */
    _handleTagClick: function _handleTagClick(e) {
      e.stopPropagation();
      this.props.changeState({
        sectionC: {
          component: 'charmbrowser',
          metadata: {
            activeComponent: 'search-results',
            search: null,
            tags: e.target.getAttribute('data-id')
          }
        }
      });
    },

    /**
      Generate the description if it is a charm.
       @method _generateDescription
      @param {Object} entityModel The entity model.
      @return {Object} The description markup.
    */
    _generateDescription: function _generateDescription(entityModel) {
      if (entityModel.get('entityType') === 'charm') {
        var bugLink = 'https://bugs.launchpad.net/charms/+source/' + ('' + entityModel.get('name'));
        var submitLink = 'https://bugs.launchpad.net/charms/+source/' + (entityModel.get('name') + '/+filebug');
        var contributeLink = 'https://code.launchpad.net/~charmers/charms/' + ('trusty/' + entityModel.get('name') + '/trunk');
        return React.createElement(
          'div',
          { className: 'row entity-content__description' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'twelve-col' },
              React.createElement(
                'p',
                null,
                entityModel.get('description')
              )
            ),
            this._generateTags(),
            React.createElement(
              'div',
              { className: 'four-col entity-content__metadata last-col' },
              React.createElement(
                'h4',
                null,
                'More information'
              ),
              React.createElement(
                'ul',
                null,
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: bugLink, target: '_blank' },
                    'Bugs'
                  )
                ),
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: submitLink, target: '_blank' },
                    'Submit a bug'
                  )
                ),
                React.createElement(
                  'li',
                  null,
                  React.createElement(
                    'a',
                    { href: contributeLink, target: '_blank' },
                    'Contribute'
                  )
                )
              )
            )
          )
        );
      }
    },

    /**
      We only show the relations when it's a charm, but not a bundle.
       @method _showEntityRelations
    */
    _showEntityRelations: function _showEntityRelations() {
      var entityModel = this.props.entityModel;
      if (entityModel.get('entityType') === 'charm') {
        return React.createElement(
          'div',
          { className: 'four-col' },
          React.createElement(juju.components.EntityContentRelations, {
            changeState: this.props.changeState,
            relations: entityModel.get('relations') })
        );
      }
    },

    /**
      We only show the revisions when it's a charm, but not a bundle.
       @method _showEntityRevisions
    */
    _showEntityRevisions: function _showEntityRevisions() {
      var entityModel = this.props.entityModel;
      if (entityModel.get('entityType') === 'charm') {
        return React.createElement(
          'div',
          { className: 'four-col' },
          React.createElement(juju.components.EntityContentRevisions, {
            revisions: entityModel.get('revisions') })
        );
      }
    },

    render: function render() {
      var entityModel = this.props.entityModel;
      return React.createElement(
        'div',
        { className: 'entity-content' },
        this._generateDescription(entityModel),
        React.createElement(
          'div',
          { className: 'row' },
          React.createElement(
            'div',
            { className: 'inner-wrapper' },
            React.createElement(
              'div',
              { className: 'seven-col append-one' },
              React.createElement(juju.components.EntityContentReadme, {
                entityModel: entityModel,
                renderMarkdown: this.props.renderMarkdown,
                getFile: this.props.getFile })
            ),
            this._showEntityRelations(),
            React.createElement(
              'div',
              { className: 'four-col' },
              React.createElement(juju.components.EntityFiles, {
                entityModel: entityModel,
                pluralize: this.props.pluralize })
            ),
            this._showEntityRevisions()
          )
        ),
        this._generateOptionsList(entityModel)
      );
    }
  });
}, '0.1.0', {
  requires: ['entity-content-config-option', 'entity-content-readme', 'entity-content-relations', 'entity-content-revisions', 'entity-files']
});
