/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('machine-view-add-machine', function () {

  juju.components.MachineViewAddMachine = React.createClass({
    displayName: 'MachineViewAddMachine',

    propTypes: {
      close: React.PropTypes.func.isRequired,
      createMachine: React.PropTypes.func.isRequired,
      machines: React.PropTypes.object,
      parentId: React.PropTypes.string,
      placeUnit: React.PropTypes.func,
      selectMachine: React.PropTypes.func,
      unit: React.PropTypes.object
    },

    /**
      Generate the initial state for the component.
       @method getInitialState
      @returns {Object} The initial state.
    */
    getInitialState: function getInitialState() {
      return {
        constraints: null,
        selectedContainer: null,
        selectedMachine: !this.props.machines && !this.props.parentId ? 'new' : null
      };
    },

    /**
      Update the state with the new constraints values.
       @method _updateConstraints
      @param {Object} constraints The new constraints values.
    */
    _updateConstraints: function _updateConstraints(constraints) {
      this.setState({ constraints: constraints });
    },

    /**
      Create the machine.
       @method _submitForm
    */
    _submitForm: function _submitForm() {
      var props = this.props;
      var state = this.state;
      var machineId = this._getParentId();
      var selectedContainer = state.selectedContainer;
      var selectedMachine = state.selectedMachine;
      var constraints = state.constraints;
      var machine;
      if (machineId && machineId !== 'new' && !selectedContainer) {
        // Don't try and create a container if the container type has not been
        // selected.
        return;
      }
      // If the state is set for a new machine or container then actually add
      // the machine/container.
      if (selectedMachine === 'new' || selectedContainer === 'lxc' || selectedContainer === 'kvm') {
        var machine = this.props.createMachine(selectedContainer, machineId, constraints);
        if (this.props.selectMachine && !machineId) {
          this.props.selectMachine(machine.id);
        }
      }
      // If the component has been provided a unit then we need to place the
      // unit on the machine/container.
      if (props.unit) {
        var id = machine && machine.id;
        props.placeUnit(props.unit, id || selectedContainer || selectedMachine);
      }
      this.props.close();
    },

    /**
      Get the parent id or selected machine.
       @method _getParentId
    */
    _getParentId: function _getParentId() {
      var selectedMachine = this.state.selectedMachine;
      if (selectedMachine === 'new') {
        return null;
      }
      return this.props.parentId || selectedMachine;
    },

    /**
      Generate the constraints form.
       @method _generateConstraints
    */
    _generateConstraints: function _generateConstraints() {
      return React.createElement(
        'div',
        { className: 'add-machine__constraints', key: 'constraints' },
        React.createElement(
          'h4',
          { className: 'add-machine__title' },
          'Define constraints'
        ),
        React.createElement(juju.components.Constraints, {
          valuesChanged: this._updateConstraints })
      );
    },

    /**
      Update the state with the selected container type.
       @method _updateSelectedContainer
      @param {Object} e The change event.
    */
    _updateSelectedContainer: function _updateSelectedContainer(e) {
      this.setState({ selectedContainer: e.currentTarget.value });
    },

    /**
      Generate the container type form.
       @method _generateSelectContainer
    */
    _generateSelectContainer: function _generateSelectContainer() {
      return React.createElement(
        'select',
        { className: 'add-machine__container',
          defaultValue: '',
          key: 'containers',
          onChange: this._updateSelectedContainer },
        React.createElement(
          'option',
          { disabled: true, value: '' },
          'Choose container type...'
        ),
        this._generateContainerOptions(),
        React.createElement(
          'option',
          { value: 'lxc' },
          'LXC'
        ),
        React.createElement(
          'option',
          { value: 'kvm' },
          'KVM'
        )
      );
    },

    /**
      Update the state with the selected machine.
       @method _updateSelectedMachine
      @param {Object} e The change event.
    */
    _updateSelectedMachine: function _updateSelectedMachine(e) {
      this.setState({ selectedMachine: e.currentTarget.value });
    },

    /**
      Generate machine selection.
       @method _generateSelectContainer
    */
    _generateSelectMachine: function _generateSelectMachine() {
      return React.createElement(
        'select',
        {
          defaultValue: '',
          key: 'machines',
          onChange: this._updateSelectedMachine },
        React.createElement(
          'option',
          { disabled: true, value: '' },
          'Move to...'
        ),
        React.createElement(
          'option',
          { value: 'new' },
          'New machine'
        ),
        this._generateMachineOptions()
      );
    },

    /**
      Generate a list of machine options.
       @method _generateMachineOptions
      @return {Array} A list of machine options.
    */
    _generateMachineOptions: function _generateMachineOptions() {
      var components = [];
      var machines = this.props.machines.filterByParent();
      machines.forEach(function (machine) {
        if (machine.deleted) {
          return;
        }
        components.push(React.createElement(
          'option',
          {
            key: machine.id,
            value: machine.id },
          machine.displayName
        ));
      });
      return components;
    },

    /**
      Generate a list of container options.
       @method _generateContainerOptions
      @return {Array} A list of container options.
    */
    _generateContainerOptions: function _generateContainerOptions() {
      var machines = this.props.machines;
      if (!machines) {
        return;
      }
      var components = [];
      var containers = machines.filterByParent(this.state.selectedMachine);
      var machineId = this._getParentId();
      components.push(React.createElement(
        'option',
        {
          key: 'root',
          value: machineId },
        machineId,
        '/root-container'
      ));
      containers.forEach(function (container) {
        if (container.deleted) {
          return;
        }
        components.push(React.createElement(
          'option',
          {
            key: container.id,
            value: container.id },
          container.displayName
        ));
      });
      return components;
    },

    /**
      Generate the buttons based on the state.
       @method _generateButtons
      @return {Object} The buttons.
    */
    _generateButtons: function _generateButtons() {
      var buttons = [{
        title: 'Cancel',
        action: this.props.close
      }, {
        title: this.props.unit ? 'Place' : 'Create',
        action: this._submitForm,
        type: 'confirm'
      }];
      return React.createElement(juju.components.ButtonRow, {
        buttons: buttons,
        key: 'buttons' });
    },

    render: function render() {
      var components = [];
      var props = this.props;
      var state = this.state;
      var machines = props.machines;
      var parentId = props.parentId;
      var selectedContainer = state.selectedContainer;
      var selectedMachine = state.selectedMachine;
      var unit = props.unit;
      if (unit && machines) {
        components.push(this._generateSelectMachine());
        if (selectedMachine) {
          if (selectedMachine !== 'new') {
            components.push(this._generateSelectContainer());
          }
          if (selectedMachine === 'new' || selectedContainer === 'kvm') {
            components.push(this._generateConstraints());
          }
          components.push(this._generateButtons());
        }
      } else if (!machines) {
        if (parentId) {
          components.push(this._generateSelectContainer());
        }
        if (!parentId || selectedContainer === 'kvm') {
          components.push(this._generateConstraints());
        }
        components.push(this._generateButtons());
      }
      return React.createElement(
        'div',
        { className: 'add-machine' },
        components
      );
    }
  });
}, '0.1.0', {
  requires: ['button-row', 'constraints']
});
