/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var juju = { components: {} }; // eslint-disable-line no-unused-vars

chai.config.includeStack = true;
chai.config.truncateThreshold = 0;

describe('NotificationList', function () {
  var _this = this;

  beforeAll(function (done) {
    // By loading this file it adds the component to the juju components.
    YUI().use('notification-list', function () {
      done();
    });
  });

  it('renders a list based on the notification passed in', function () {
    var notification = {
      timestamp: '12345',
      message: 'notification message',
      level: 'info'
    };
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.NotificationList, {
      notification: notification }));
    var children = output.props.children;
    var items = [React.createElement(juju.components.NotificationListItem, {
      key: notification.timestamp,
      timestamp: notification.timestamp,
      ref: 'NotificationListItem' + notification.timestamp,
      removeNotification: children[0].props.removeNotification,
      message: notification.message,
      timeout: undefined,
      type: notification.level })];
    var expected = React.createElement(
      'ul',
      { className: 'notification-list' },
      items
    );
    assert.deepEqual(output, expected);
  });

  it('can render with no notifications', function () {
    var output = jsTestUtils.shallowRender(React.createElement(juju.components.NotificationList, {
      notification: null }));
    assert.deepEqual(output.props.children[0], undefined);
  });

  it('can render notifications after rendering none', function () {
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.NotificationList, {
      notification: null }), true);
    var output = renderer.getRenderOutput();
    assert.deepEqual(output.props.children[0], undefined);
    var notification = {
      timestamp: '12345',
      message: 'notification message',
      level: 'info'
    };
    renderer.render(React.createElement(juju.components.NotificationList, {
      notification: notification }));
    var instance = renderer.getMountedInstance();
    // Updates to state don't re-render when using the shallow renderer
    // so just checking that the state gets updated.
    assert.deepEqual(instance.state, {
      notifications: {
        12345: {
          message: 'notification message',
          type: 'info'
        }
      }
    });
  });

  it('times out non error messages', function (done) {
    var notification = {
      timestamp: '12345',
      message: 'notification message',
      level: 'info'
    };
    var timeout = 1;
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.NotificationList, {
      timeout: timeout,
      notification: notification }), _this);
    var instance = renderer.getMountedInstance();
    // Fake the ref being created for the list item;
    instance.refs = {
      NotificationListItem12345: {
        hide: function hide() {
          done();
        }
      }
    };
  });

  it('does not time out error messages', function () {
    var notification = {
      timestamp: '12345',
      message: 'notification message',
      level: 'error'
    };
    var timeout = 1;
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.NotificationList, {
      timeout: timeout,
      notification: notification }), _this);
    var instance = renderer.getMountedInstance();
    // Fake the ref being created for the list item;
    instance.refs = {
      NotificationListItem12345: {
        hide: function hide() {
          // If it gets here then it's setting the timeout.
          assert.fail();
        }
      }
    };
  });
});
