/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2015 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('unit-list', function () {

  juju.components.UnitList = React.createClass({
    displayName: 'UnitList',

    /**
      Fires changeState to update the UI based on the component clicked.
       @method _navigate
      @param {Object} e The click event.
    */
    _navigate: function _navigate(e) {
      this.props.changeState({
        sectionA: {
          component: 'inspector',
          metadata: {
            id: this.props.service.get('id'),
            activeComponent: 'scale'
          } } });
    },

    /**
      Sets the selectAll state property based on the "select all" child
      component.
       @method _selectAllUnits
      @param {String} group The key for a group of checkboxes.
      @param {Boolean} checked Whether the "select all" child component is
        checked.
    */
    _selectAllUnits: function _selectAllUnits(group, checked) {
      var _this = this;

      var setChecked = function setChecked(key, groups) {
        groups[key].units.forEach(function (unit) {
          _this.refs['UnitListItem-' + unit.id].setState({
            checked: checked
          });
        });
      };
      if (checked === undefined) {
        checked = false;
      }
      var groups = this._generateGroups();
      if (group === null) {
        for (var key in groups) {
          setChecked(key, groups);
        }
      } else {
        setChecked(group, groups);
      }
    },

    /**
      The callable to be passed to the unit items for navigating to the unit
      details.
       @method _unitItemAction
      @param {Object} e The click event.
    */
    _unitItemAction: function _unitItemAction(e) {
      var unitParts = e.currentTarget.getAttribute('data-id').split('/');
      this.props.changeState({
        sectionA: {
          component: 'inspector',
          metadata: {
            // This is done in parts like this because subordinate units show
            // the service unit that it's placed on, not the subordinate itself.
            id: unitParts[0], // Service Id
            unit: unitParts[1], // Unit Id
            activeComponent: 'unit'
          }
        }
      });
    },

    /**
      Update the selected units with the supplied action.
       @method _handleUpdateUnits
      @param {String} action The action to apply to the units.
    */
    _handleUpdateUnits: function _handleUpdateUnits(action) {
      var units = [];
      var refs = this.refs;
      var envResolved = this.props.envResolved;
      Object.keys(refs).forEach(function (ref) {
        var isInstance = ref.split('-')[0] === 'UnitListItem';
        if (isInstance && refs[ref].state.checked) {
          var unitName = ref.slice(ref.indexOf('-') + 1);
          units.push(unitName);
          // On resolve and remove we want to mark the unit as resolved else
          // Juju won't remove units that are in error.
          if (action === 'resolve' || action === 'remove') {
            envResolved(unitName, null, false);
          } else if (action === 'retry') {
            envResolved(unitName, null, true);
          }
        }
      });
      if (action === 'remove') {
        this.props.destroyUnits(units);
      }
      this._selectAllUnits(null, false);
    },

    /**
      Generates a list of unit components.
       @method _generateUnitList
      @param {Object} group A definition for a group of checkboxes .
      @returns {Array} Collection of unit components.
    */
    _generateUnitList: function _generateUnitList(group) {
      var _this2 = this;

      var key = group.key;
      var unitList = [React.createElement(juju.components.UnitListItem, {
        key: key,
        ref: key,
        label: group.label,
        className: 'select-all',
        whenChanged: this._selectAllUnits.bind(this, key) })];
      group.units.forEach(function (unit) {
        var ref = 'UnitListItem-' + unit.id;
        unitList.push(React.createElement(juju.components.UnitListItem, {
          key: unit.displayName,
          ref: ref,
          label: unit.displayName,
          action: _this2._unitItemAction,
          unitId: unit.id }));
      });
      return unitList;
    },

    /**
      Generate the groups of units for the service.
       @method _generateGroups
      @returns {Object} The groups of units for the service.
    */
    _generateGroups: function _generateGroups() {
      var units = this.props.units;
      var groups = {};
      if (this.props.unitStatus === 'error') {
        var errors = {};
        units.forEach(function (unit) {
          var agentState = unit.agent_state_info;
          if (!errors[agentState]) {
            errors[agentState] = [];
          }
          errors[agentState].push(unit);
        });
        Object.keys(errors).forEach(function (error, i) {
          var key = 'select-all-' + i;
          groups[key] = {
            label: error,
            units: errors[error],
            key: key
          };
        });
      } else {
        var key = 'select-all';
        groups[key] = {
          label: 'Select all units',
          units: units,
          key: key
        };
      }
      return groups;
    },

    /**
      Generate the groups of units.
       @returns {Object} The list components
    */
    _generateListGroups: function _generateListGroups() {
      var components = [];
      var groups = this._generateGroups();
      Object.keys(groups).forEach(function (key) {
        var group = groups[key];
        components = components.concat(this._generateUnitList(group));
      }, this);
      return components;
    },

    /**
      Generate the buttons for the status.
       @returns {Array} The list of buttons
    */
    _generateButtons: function _generateButtons() {
      var buttons = [];
      if (this.props.unitStatus === 'error') {
        buttons.push({
          title: 'Resolve',
          action: this._handleUpdateUnits.bind(this, 'resolve')
        });
        buttons.push({
          title: 'Retry',
          action: this._handleUpdateUnits.bind(this, 'retry')
        });
      }
      buttons.push({
        title: 'Remove',
        action: this._handleUpdateUnits.bind(this, 'remove')
      });
      return buttons;
    },

    /**
      Generate the scale service action.
       @returns {Object} The scale service component.
    */
    _generateScaleService: function _generateScaleService() {
      // Don't show the scale service if we're viewing a status list (e.g.
      // errors) or if the service is a subordinate.
      if (this.props.unitStatus || this.props.service.get('subordinate')) {
        return;
      }
      return React.createElement(
        'div',
        { className: 'unit-list__actions' },
        React.createElement(juju.components.OverviewAction, {
          action: this._navigate,
          icon: 'plus_box_16',
          title: 'Scale service' })
      );
    },

    render: function render() {
      return React.createElement(
        'div',
        { className: 'unit-list' },
        this._generateScaleService(),
        React.createElement(
          'ul',
          { className: 'unit-list__units' },
          this._generateListGroups()
        ),
        React.createElement(juju.components.ButtonRow, {
          buttons: this._generateButtons() })
      );
    }

  });
}, '0.1.0', { requires: ['overview-action', 'button-row', 'unit-list-item', 'overview-action'] });
