# Copyright 2015 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import json
import logging
import os
import pkg_resources

from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import FileResponse
from pyramid.view import view_config


ASSET_PATH = os.path.join(
    os.path.dirname(__file__),
    'static',
    'gui',
    'build',
    'app'
    )


VERSION = pkg_resources.get_distribution('jujugui').version
log = logging.getLogger('jujugui')


@view_config(
    route_name='jujugui.app',
    renderer='jujugui:templates/index.html.mako',
    http_cache=3600 * 24 * 365)
def app(request):
    """The main Juju GUI JavaScript application."""
    env_uuid = request.matchdict.get('uuid')
    settings = request.registry.settings
    cache_buster = settings.get('jujugui.cachebuster', VERSION)

    return {
        'config_url': request.route_path('jujugui.config', uuid=env_uuid),
        'convoy_url': request.route_path(
            'jujugui.convoy', cachebuster=cache_buster, uuid=env_uuid),
        'combine': settings['jujugui.combine'],
        'raw': settings['jujugui.raw'],
    }


@view_config(route_name='jujugui.ui')
def juju_ui(request):
    requested_file = request.matchdict.get('file')
    file_path = os.path.join(ASSET_PATH, requested_file)
    try:
        return FileResponse(file_path, request=request)
    except OSError:
        log.info('No such file: %s' % file_path)
        return HTTPNotFound()


@view_config(route_name='jujugui.config', renderer='string')
def config(request):
    """The Juju GUI dynamically generated config.js file.

    This file includes the GUI configuration options, including the Juju API
    configuration.
    """
    settings = request.registry.settings
    request.response.content_type = 'application/javascript'
    sandbox_enabled = settings['jujugui.sandbox']
    jem_url = settings['jujugui.jem_url']
    # If sandbox is enabled then set the password to "admin" so that the
    # Juju GUI will automatically log in.
    user, password = 'admin', 'password'
    if not sandbox_enabled:
        user, password = settings['jujugui.user'], settings['jujugui.password']
    baseUrl = settings['jujugui.base_url']
    env_uuid = settings.get('jujugui.uuid')
    path_env_uuid = request.matchdict.get('uuid')
    if baseUrl is None:
        if path_env_uuid is None:
            baseUrl = ''
        else:
            baseUrl = '/u/anonymous/{}'.format(path_env_uuid)
    env_uuid = [
        uuid for uuid in [path_env_uuid, env_uuid, 'sandbox']
        if uuid is not None][0]
    options = {
        # Base YUI options.
        'auth': settings['jujugui.auth'],
        'serverRouting': False,
        'html5': True,
        'container': '#main',
        'viewContainer': '#main',
        'baseUrl': baseUrl,
        'transitions': False,
        'cachedFonts': False,
        # Debugging options.
        'consoleEnabled': True,
        'simulateEvents': False,
        # The charm store URL and path.
        'charmstoreURL': settings['jujugui.charmstore_url'],
        'apiPath': settings['jujugui.api_path'],
        # WebSocket connection to the Juju API.
        'socket_protocol': 'wss',
        'user': user,
        'password': password,
        'jujuEnvUUID': env_uuid,
        'jemUrl': jem_url,
        'interactiveLogin': settings['jujugui.interactive_login'],
        # Enable/disable sandbox (demonstration) mode.
        'sandbox': sandbox_enabled,
        'sandboxSocketURL': 'wss://demo.jujucharms.com/ws',
        # XXX frankban: do we still support read-only mode?
        'readOnly': False,
        # Set the GA_key to enable Google Analytics usage and calls.
        # Also implies using cookies.
        'GA_key': settings['jujugui.ga_key'],
        # Set a juju-core version so the GUI can adapt its available features.
        'jujuCoreVersion': settings.get('jujugui.jujuCoreVersion', ''),
        'apiAddress': settings.get('jujugui.apiAddress'),
        'socketTemplate': settings['jujugui.socketTemplate'],
    }
    return 'var juju_config = {};'.format(json.dumps(options))


@view_config(route_name='jujugui.version', renderer='prettyjson')
def version(request):
    return {'version': VERSION}
